/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

rw.c is for reading from one file, (which may be a named pipe,) and
writing what is read to another file, (which may be a named pipe.)

It is a template/skeleton for constructing formal file operations on
Unix, (specifically for daemons that listen on a pipe.)

    1) If the input file is a named pipe, it is opened with O_RDWR,
       (with no locks,) and if the process writing to the pipe exits,
       this program will continue to "listen" on the pipe for another
       process to open the pipe and write data.

    2) The output file is opened for each input record from the input
       file-and closed after the record is written to the output file.

       Note that opening and closing the output file for every record
       is a time consuming process-unless absolutely necessary, it is
       advisable to open and close the output file at the beginning
       and end of the program.

    3) A mandatory lock is placed on the output file, unless it is
       stdout.

       If acquisition of the lock fails, an attempt will be made in
       one second, (up to lock_wait many seconds,) for the lock.

    4) If the output file is a named pipe, or stdout, it is not
       flushed; if it is a regular file, it is.

       Note that flushing the output file after every record written
       is a time consuming process-unless flushing is absolutely
       necessary, it is advisable to remove the write flush code.

       Flushing a named pipe causes the listing program to exit,
       and stdout should not be flushed.

    5) The input file is read, one character at a time, and the
       buffered record processed, (written to the output file in this
       case,) on any '\n' character. The read (2) function may not be
       considered canonical in all cases, (and partial records may be
       returned from read (2)).

$Revision: 0.0 $
$Date: 2007/12/18 19:36:00 $
$Id: rw.c,v 0.0 2007/12/18 19:36:00 john Exp $
$Log: rw.c,v $
Revision 0.0  2007/12/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#define BUFLEN BUFSIZ /* size of record buffers */

static char rcsid[] = "$Id: rw.c,v 1.12 12/01/07 21:31:06 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Read from a file, write to a file\n",
    "Usage: rw [-i infilename] [-o outfilename] [-v]\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -o outfilename, output file name, (stdout)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening input file\n",
    "Error closing input file\n",
    "Error opening output file\n",
    "Error closing output file\n",
    "Error locking output file\n",
    "Error unlocking output file\n",
    "Error reading input file\n",
    "Error writing to output file\n",
    "Error syncing output file\n",
    "Error overflow buffer\n",
    "Error getting status of input file\n",
    "Error getting status of output file\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EIOPEN 2
#define EICLOSE 3
#define EOOPEN 4
#define EOCLOSE 5
#define EOLCK 6
#define EOULCK 7
#define EIREAD 8
#define EOWRITE 9
#define EOSYNC 10
#define EOVRFL 11
#define EISTAT 12
#define EOSTAT 13

enum file_type /* input and output file type */
{
    REG, /* regular file */
    PIPE, /* pipe */
    TTY /* terminal, default */
};

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */

#else

static void print_message (); /* print any error messages */

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char *in_name = (char *) 0, /* input file name */
         *out_name = (char *) 0, /* output file name */
         inchar, /* character from input file */
         inbuffer[BUFLEN + 1], /* input read buffer */
         *ref, /* character reference in input buffer */
         *sentinel = &inbuffer[BUFLEN]; /* reference to sentinel character in input buffer */

    int retval = NOERROR, /* return value, assume no error */
        infile = STDIN_FILENO, /* input file descriptor */
        outfile = STDOUT_FILENO, /* output file descriptor */
        input_mode = O_RDONLY, /* open mode for input file */
        lock_wait = 10, /* maximum number of seconds to wait for an exclusive lock on the output file before an error is issued, and the program exits */
        c; /* command line switch */

    ssize_t cnt_in = 0, /* count of characters read */
            cnt_out; /* count of characters to write */

    struct flock lockfile; /* lock file structure */

    struct stat file_stat; /* lstat () structure */

    enum file_type in_type = TTY, /* input file type */
                   out_type = TTY; /* output file type */

    *sentinel = '\0'; /* sentinel character in input buffer */

    while ((retval == NOERROR) && ((c = getopt (argc, argv, "i:o:v")) != EOF)) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                retval = EISTAT; /* assume error getting status of input file */

                if (lstat (in_name, &file_stat) == 0) /* get the input file status */
                {

                    if (S_ISFIFO (file_stat.st_mode)) /* input file a pipe? */
                    {
                        in_type = PIPE;
                        input_mode = O_RDWR; /* open mode for input file */
                    }

                    else if (S_ISREG (file_stat.st_mode)) /* input file a regular file? */
                    {
                        in_type = REG;
                    }

                    /* else the input file type is defaulted to TTY */

                    retval = NOERROR; /* assume no error */
                }

                break;

            case 'o': /* request for output file name? */

                out_name = optarg; /* yes, save the output file name */
                retval = EOSTAT; /* assume error getting status of output file */

                if (lstat (out_name, &file_stat) == 0) /* get the output file status */
                {

                    if (S_ISFIFO (file_stat.st_mode)) /* output file a pipe? */
                    {
                        out_type = PIPE;
                    }

                    else if (S_ISREG (file_stat.st_mode)) /* output file a regular file? */
                    {
                        out_type = REG;
                    }

                    /* else the output file type is defaulted to TTY */

                    retval = NOERROR; /* assume no error */
                }

                else if (errno == ENOENT) /* a non-existent output file will be created by open () as a regular file, below; any other lstat () error is a hard error */
                {
                    out_type = REG;
                    retval = NOERROR; /* assume no error */
                }

                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EIOPEN; /* assume error opening input file */

        if ((infile = (in_name == (char *) 0) ? infile : open (in_name, input_mode)) >= 0) /* open the input file */
        {
            ref = inbuffer; /* reset the character reference in input buffer */
            *ref = '\0'; /* terminate the record */
            retval = EIREAD; /* assume error reading input file */

            /* retval == EIREAD means no error in progress; any other error means to drop through with the error, (without reading the next record); an error in reading from the input file means drop through with the read error */

            while ((retval == EIREAD) && ((cnt_in = read (infile, &inchar, sizeof (char)) == sizeof (char)))) /* read characters from the input file */
            {
                retval = EOVRFL; /* assume error overflow buffer */

                if (ref < sentinel) /* buffer overflow? */
                {
                    *ref = inchar; /* save the character in the input buffer */
                    ref ++; /* prepare for next character reference in input buffer */
                    *ref = '\0'; /* terminate the record; may overwrite the sentinel '\0' */
                    retval = EIREAD; /* assume error reading input file */

                    if (inchar == '\n') /* character an EOL? */
                    {
                        cnt_out = ref - inbuffer; /* save the count of characters read */
                        retval = EOOPEN; /* assume error opening output file */

                        if ((outfile = (out_name == (char *) 0) ? outfile : open (out_name, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH)) >= 0) /* open the output file */
                        {
                            lockfile.l_type = F_WRLCK; /* exclusive write lock the entire output file */
                            lockfile.l_whence = (short int) SEEK_SET;
                            lockfile.l_start = (off_t) 0;
                            lockfile.l_len = (off_t) 0;
                            retval = EOLCK; /* assume error opening output file */

                            while (retval == EOLCK) /* wait to obtain a lock on the output file; if the maximum time is exceeded, a break will be issued below to break out of this loop */
                            {

                                if ((out_type == TTY) || (fcntl (outfile, F_SETLK, &lockfile) == 0)) /* exclusive write lock the output file, but only if its a regular file or pipe */
                                {
                                    retval = EOWRITE; /* assume error writing to output file */

                                    if ((write (outfile, inbuffer, (size_t) cnt_out)) == cnt_out) /* write the characters to the output device */
                                    {
                                        retval = EIREAD; /* assume error reading input file */

                                        if ((out_type == REG) && (fsync (outfile) != 0)) /* flush the output file's buffers, but only if its a regular file; do not flush pipes or ttys */
                                        {
                                            retval = EOSYNC; /* assume error syncing output file */
                                        }

                                    }

                                    lockfile.l_type = F_UNLCK; /* unlock the file */

                                    if ((out_type != TTY) && (fcntl (outfile, F_SETLKW, &lockfile)) != 0) /* unlock the output file, but only if its a regular file or pipe */
                                    {
                                        retval = (retval == NOERROR) ? EOULCK : retval; /* assume error unlocking output file */
                                    }

                                }

                                else /* locking the output file failed, wait one second, and try again up to loackwait many seconds */
                                {
                                    sleep (1); /* sleep one second, for lock_wait many times */

                                    if (-- lock_wait == 0) /* waited to obtain a lock on the output file long enough?; if the maximum time is exceeded, a break will be issued to break out of this loop */
                                    {
                                        break; /* timed out, break out of this program */
                                    }

                                }

                            }

                            if ((outfile != STDOUT_FILENO) && (close (outfile) != 0)) /* close the output file, but only if it is not stdout */
                            {
                                retval = (retval == NOERROR) ? EOCLOSE : retval; /* assume error closing output file */
                            }

                        }

                        ref = inbuffer; /* reset the character reference in input buffer */
                        *ref = '\0'; /* terminate the record */
                    }

                }

            }

            retval = ((cnt_in == 0) && (retval == EIREAD)) ? NOERROR : retval; /* a count of zero, and error reading input file was an EOF or ^D, a normal exit */

            if ((infile != STDIN_FILENO) && (close (infile) != 0)) /* close the input file, but only if its not stdin */
            {
                retval = (retval == NOERROR) ? EICLOSE : retval; /* assume error closing input file */
            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
